@@ -1,12 +1,10 @@
 # set encoding=utf-8
 
 import sys
+import os
 import time
 import inspect
 import re
-import logging
-import traceback
-import os
 import linecache
 from pprint import pformat
 try:
@@ -16,7 +14,7 @@ except ImportError:
 import __builtin__
 
 
-__version__ = '1.2.2'
+__version__ = '1.2.3'
 __author__ = 'Yoshida Tetsuya'
 __license__ = 'MIT License'
 
@@ -43,12 +41,12 @@ class PyXdebug(object):
         self.end_gmtime = None
         self.call_depth = 0
         self.call_func_name = None
-        self.late_dispatch = None
+        self.late_dispatch = []
         self.result = []
 
     def run_func(self, func, *args, **kwds):
         self.initialize()
-        self.call_func_name = func.__name__
+        self.call_func_name = getattr(func, '__name__', None)
         return self._run(func, *args, **kwds)
 
     def run_statement(self, statement):
@@ -111,11 +109,6 @@ class PyXdebug(object):
             # call
             return func(*args, **kwds)
         finally:
-            # late
-            if self.late_dispatch:
-                self.late_dispatch()
-                self.late_dispatch = None
-
             # reset profile hook
             sys.settrace(original_trace)
 
@@ -127,7 +120,7 @@ class PyXdebug(object):
                 __builtin__.__import__ = original_import
                 __builtin__.reload = original_reload
 
-            # end
+            # finish
             trace = FinishTrace(None, 0)
             trace.setvalue(self.start_time)
             self.result.append(trace)
@@ -142,25 +135,41 @@ class PyXdebug(object):
             if self.call_func_name is None or self.call_func_name!=frame.f_code.co_name:
                 return
 
-        # late
-        if self.late_dispatch:
-            self.late_dispatch()
-            self.late_dispatch = None
+        # normalize event
+        if event[0:2]=='c_':
+            event = event[2:]
 
-        # dispatch
-        if event=='call' or event=='c_call':
+        # wrap frame
+        frame = FrameWrap(frame)
+        frame.f_back = FrameWrap(frame.f_back)
+
+        print "%-7s %-3d %s" % (event, frame.f_lineno, frame.get_line().rstrip())
+
+        # dispatch call
+        if event=='call':
             self.trace_call(frame, arg)
-        elif event=='return' or event=='c_return':
+
+            # collect assignments
+            if self.collect_assignments:
+                self.late_dispatch.append(None)
+
+        # dispatch return
+        elif event=='return':
+            # collect assignments
+            if self.collect_assignments:
+                self.trace_line(frame, arg)
+                self.late_dispatch.pop()
+
             self.trace_return(frame, arg)
-        elif event=='line' and self.collect_assignments:
-            def late():
+
+        # dispatch line
+        elif event=='line':
+            if self.collect_assignments:
                 self.trace_line(frame, arg)
-            self.late_dispatch = late
 
         return self.trace_dispatch
 
     def trace_call(self, frame, arg):
-        trace_index = len(self.result)
         trace = CallTrace(frame, self.call_depth)
         trace.setvalue(self.start_time, self.collect_params)
         self.result.append(trace)
@@ -174,11 +183,22 @@ class PyXdebug(object):
             self.result.append(trace)
 
     def trace_line(self, frame, arg):
-        filename = frame.f_code.co_filename
-        lineno = frame.f_lineno
-        line = linecache.getline(filename, lineno).strip()
+        back_frame = self.late_dispatch[self.call_depth-1]
+        self.late_dispatch[self.call_depth-1] = frame
+
+        if back_frame:
+            frame = FrameWrap(frame)
+            frame.set_position(back_frame)
+            self._trace_line(frame, None)
+
+    def _trace_line(self, frame, arg):
+        line = frame.get_line().strip()
         match = re.compile(r"^([^\+\-\*/=]+)([\+\-\*/]?=[^=])(.+)$").match(line)
         if match:
+            log = LogTrace(frame, self.call_depth)
+            log.setvalue('%d execlate: %s' % (frame.f_lineno, frame.f_locals))
+            #self.result.append(log)
+
             varnames = match.group(1).strip()
             try:
                 varnames = re.compile(r"^\((.+)\)$").match(varnames).group(1).strip()
@@ -190,24 +210,16 @@ class PyXdebug(object):
                 varnames = None
             if varnames:
                 for varname in varnames:
-                    objectname = None
-                    attrname = None
-                    value = None
-                    try:
-                        objectname, attrname = varname.split('.')
-                    except:
-                        pass
-                    if objectname:
-                        object = frame.f_locals.get(objectname, None)
-                        value = getattr(object, attrname, None)
-                    else:
-                        value = frame.f_locals.get(varname, None)
+                    value = get_frame_var(frame, varname)
                     trace = SubstituteTrace(frame, self.call_depth)
                     trace.setvalue(varname, value)
                     self.result.append(trace)
+        else:
+            log = LogTrace(frame, self.call_depth)
+            log.setvalue('nomatch: %s %s' % (line, frame.f_locals))
+            #self.result.append(log)
 
     def trace_import(self, frame, arg):
-        trace_index = len(self.result)
         trace = ImportTrace(frame, self.call_depth)
         trace.setvalue(arg[0], arg[1], self.start_time)
         self.result.append(trace)
@@ -372,6 +384,19 @@ class FinishTrace(CallTrace):
         return u'%10.4f %10d' % (self.time or 0.0, self.memory or 0)
 
 
+class LogTrace(AbstractTrace):
+    def __init__(self, callee, call_depth):
+        super(LogTrace, self).__init__(callee, call_depth)
+        self.message = None
+
+    def setvalue(self, message):
+        self.message = message
+
+    def get_result(self):
+        sp = u' '*24 + u'  '*self.call_depth
+        return u'%s*> %s' % (sp, self.message)
+
+
 class PyXdebugError(Exception):
     pass
 
@@ -421,6 +446,64 @@ def get_method_name(frame):
     return methodname
 
 
+def get_frame_var(frame, varname):
+    objectname = None
+    attrname = None
+    value = None
+    try:
+        objectname, attrname = varname.split('.')
+    except:
+        pass
+    if objectname:
+        object = frame.f_locals.get(objectname, None)
+        value = getattr(object, attrname, None)
+    else:
+        value = frame.f_locals.get(varname, None)
+    return value
+
+
+def is_substitute(frame):
+    line = frame.get_line().strip()
+    if re.compile(r"^([^\+\-\*/=]+)([\+\-\*/]?=[^=])(.+)$").match(line):
+        return True
+    return False
+
+
+def is_return(frame):
+    line = frame.get_line().strip()
+    words = re.compile(r"\s+").split(line)
+    if len(words) and words[0]=='return':
+        return True
+    return False
+
+
+class FrameWrap(object):
+    def __init__(self, frame):
+        keys = ('f_back', 'f_builtins', 'f_code', 'f_exc_traceback', 'f_exc_type', 'f_exc_value', 'f_globals', 'f_lasti', 'f_lineno', 'f_locals', 'f_restricted', 'f_trace')
+        for key in keys:
+            value = getattr(frame, key, None)
+            setattr(self, key, value)
+
+    def get_line(self):
+        filename = self.f_code.co_filename
+        lineno = self.f_lineno
+        return linecache.getline(filename, lineno)
+
+    def set_position(self, other):
+        self.f_code = getattr(other, 'f_code', None)
+        self.f_lineno = getattr(other, 'f_lineno', None)
+
+    def is_equal_position(self, other):
+        try:
+            lineno1 = self.f_lineno
+            filename1 = self.f_code.co_filename
+            lineno2 = other.f_lineno
+            filename2 = other.f_code.co_filename
+            return lineno1==lineno2 and filename1==filename2
+        except:
+            return False
+
+
 #=================================================
 
 
